// ROOT Libraries
#include "TCanvas.h"
#include "TEntryList.h"
#include "TFile.h"
#include "THStack.h"
#include "TLegend.h"
#include "TPaveStats.h"
#include "TROOT.h"
#include "TStyle.h"
// Standard Libraries
#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <map>
#include <sstream>
// Other
#include "createplot.h"

using std::cin;
using std::cout;
using std::endl;
using std::fixed;
using std::flush;
using std::ifstream;
using std::map;
using std::scientific;
using std::setprecision;
using std::setw;
using std::stringstream;

void makePlots(vector<HS> histograms, DrawingOptions drawOp, string outputDir) {
	bool logScale = true;
	bool setMin = true;
	bool setMax = true;

	if (!histograms.size()) {
		cout << "WARNING: No histograms to plot!" << endl;
		return;
	}

	gROOT->SetStyle( "Plain" );
	gStyle->SetOptStat( "" );
	TCanvas * canv = new TCanvas( "HZZ2l2nu", "HZZ2l2nu" );

	TLegend * leg = new TLegend( 0.7, 0.8, 0.95, 1.00 );
	double font_size = (( 0.2 / histograms.size() ) < 0.04 ) ? ( 0.2 / histograms.size() ) : 0.04;
	leg->SetTextSize( font_size );
	leg->SetFillColor( 0 );

	string stackName(histograms[0].first->GetName());
	THStack histoStack( "HistogramsStack", stackName.c_str() );
	for (unsigned i = 0; i < histograms.size(); ++i) {
		unsigned stack = drawOp.getStack( histograms[i].second );
		if (stack) {
			drawOp.applyOptions(histograms[i].first, histograms[i].second);
			Option_t * legOpt = drawOp.getLegendOptions(histograms[i].second).c_str();
			leg->AddEntry(histograms[i].first, histograms[i].second.c_str(), legOpt );
			histoStack.Add( histograms[i].first );
		}
	}
	if (!histoStack.GetStack())
		cout << "WARNING: Empty stack!" << endl;

	vector<HS> histos;
	for (unsigned i = 0; i < histograms.size(); ++i) {
		DrawingOptions::TYPE histType = drawOp.getHistogramType(histograms[i].second);
		unsigned stack = drawOp.getStack( histograms[i].second );
		if (!stack && histType == DrawingOptions::FILL )
			histos.push_back( histograms[i] );
	}
	for (unsigned i = 0; i < histograms.size(); ++i) {
		DrawingOptions::TYPE histType = drawOp.getHistogramType(histograms[i].second);
		unsigned stack = drawOp.getStack( histograms[i].second );
		if (!stack && histType == DrawingOptions::HATCH )
			histos.push_back( histograms[i] );
	}
	for (unsigned i = 0; i < histograms.size(); ++i) {
		DrawingOptions::TYPE histType = drawOp.getHistogramType(histograms[i].second);
		unsigned stack = drawOp.getStack( histograms[i].second );
		if (!stack && histType == DrawingOptions::MARK )
			histos.push_back( histograms[i] );
	}

	vector<TH1D *> histosP;
	for (unsigned i = 0; i < histograms.size(); ++i)
		histosP.push_back(histograms[i].first);
	const double min = findHistosNonZeroMinimum( histosP );
	const double max = findHistosMaximum( histosP, histoStack );

	canv->cd();
	if (logScale)
		gPad->SetLogy( 1 );
	else
		gPad->SetLogy( 0 );
	gPad->Update();

	bool same = false;
	
	if ( histoStack.GetStack() ) {
		if (setMin && logScale && min > 0)
			histoStack.SetMinimum( min );
		if (setMin && !logScale)
			histoStack.SetMinimum( 0 );
		if (setMax && logScale)
			histoStack.SetMaximum( 2 * max);
		if (setMax && !logScale)
			histoStack.SetMaximum( 1.1 * max);
		histoStack.Draw("");
		gPad->Update();
		same = true;
	}
	for (unsigned i = 0; i < histos.size(); ++i) {
		if (setMin && logScale && min > 0)
			histos[i].first->SetMinimum( min );
		if (setMin && !logScale)
			histos[i].first->SetMinimum( 0 );
		if (setMax && logScale)
			histos[i].first->SetMaximum( 2 * max);
		if (setMax && !logScale)
			histos[i].first->SetMaximum( 1.1 * max);
		drawOp.applyOptions(histos[i].first, histos[i].second);
		histos[i].first->Print();
		Option_t * drawOpt = drawOp.getDrawOptions(histos[i].second, same).c_str();
		histos[i].first->Draw( drawOpt );
		Option_t * legOpt = drawOp.getLegendOptions(histos[i].second).c_str();
		leg->AddEntry(histos[i].first, histos[i].second.c_str(), legOpt );
		gPad->Update();
		same = true;
	}
	leg->Draw( "same" );
	gPad->RedrawAxis();

	string mkdir = "mkdir -p " + outputDir;
	if( system( mkdir.c_str() ) )
		cout << "Problem with creating directory!" << endl;

	string fileName = outputDir + histos[0].first->GetName() + ".ps";
	canv->SaveAs( fileName.c_str() );

	delete canv;
	delete leg;
}

DrawingOptions::DrawingOptions(const char * fn) {
	readOptions(fn);
}

void DrawingOptions::readOptions(const char * fileName) {
	ifstream input( fileName );
	if( !input.is_open() ) {
		cout << "ERROR: Can't open file with options (" << fileName << ")!" << endl;
		exit( EXIT_FAILURE );
	}
	while( !input.eof() ) {
		string temp;
		getline( input, temp );
		if( !temp.size() || temp[0] == '\n' || temp[0] == '#' ) {
			continue;
		}
		HistOpt tempHO;	
		stringstream in( temp );
		char tempCH;
		if (temp[0] == 'F') {
			in >> tempCH >> tempHO.name_ >> tempHO.size_ >> tempHO.color_ >> tempHO.leg_ >> tempHO.stack_;
			tempHO.type_ = FILL;
		} else if (temp[0] == 'H') {
			in >> tempCH >> tempHO.name_ >> tempHO.style_ >> tempHO.size_ >> tempHO.color_ >> tempHO.leg_ >> tempHO.stack_;
			tempHO.type_ = HATCH;
		} else if (temp[0] == 'M') {
			in >> tempCH >> tempHO.name_ >> tempHO.style_ >> tempHO.size_ >> tempHO.color_ >> tempHO.leg_ >> tempHO.draw_ >> tempHO.stack_;
			tempHO.type_ = MARK;
		} else {
			cout << "ERROR: Unknown type of options!" << endl;
			exit( EXIT_FAILURE );
		}
		options_.push_back(tempHO);
	}
}

void DrawingOptions::applyOptions(TH1D * histo, string name) {
	for (unsigned i = 0; i < options_.size(); ++i) {
		if ( options_[i].name_ == name ) {
			if (options_[i].type_ == MARK) {
				histo->SetMarkerStyle( options_[i].style_ );
				histo->SetMarkerSize( options_[i].size_ );
				histo->SetMarkerColor( options_[i].color_ );
			} else if (options_[i].type_ == FILL) {
				histo->SetLineWidth( options_[i].size_ );
				histo->SetFillColor( options_[i].color_ );
			} else if (options_[i].type_ == HATCH) {
				gStyle->SetHatchesSpacing( 1.0 );
				gStyle->SetHatchesLineWidth( 3 );
				histo->SetFillStyle( options_[i].style_ );
				histo->SetLineWidth( options_[i].size_ );
				histo->SetLineColor( options_[i].color_ );
				histo->SetFillColor( options_[i].color_ );
			} else {
				cout << "ERROR: Type not known!" << endl;
				exit( EXIT_FAILURE );
			}
			break;
		}
	}
}

string DrawingOptions::getLegendOptions(const string & name) {
	for (unsigned i = 0; i < options_.size(); ++i) {
		if ( options_[i].name_ == name ) {
			return options_[i].leg_;
		}
	}
	return string("");
}

string DrawingOptions::getDrawOptions(const string & name, bool same) {
	for (unsigned i = 0; i < options_.size(); ++i) {
		if ( options_[i].name_ == name ) {
			string temp;
			if (same)
				temp = "same" + options_[i].draw_;
			else
				temp = options_[i].draw_;
			return temp;
		}
	}
	return string("");
}

DrawingOptions::TYPE DrawingOptions::getHistogramType(const string & name) {
	for (unsigned i = 0; i < options_.size(); ++i) {
		if ( options_[i].name_ == name ) {
			return options_[i].type_;
		}
	}
	return TYPE(FILL);
}

unsigned DrawingOptions::getStack(const string & name) {
	for (unsigned i = 0; i < options_.size(); ++i) {
		if ( options_[i].name_ == name ) {
			return options_[i].stack_;
		}
	}
	return 0;
}

double findHistosNonZeroMinimum( const vector<TH1D *> histos ) {
	double totMin = 0;
	double min;
	for (unsigned j = 0; j < histos.size(); ++j) {
		int nBins = histos[j]->GetNbinsX();
		int i = 1;
		min = 0;
		while( i <= nBins ) {
			if( histos[j]->GetBinContent( i ) > 0 ) {
				min = histos[j]->GetBinContent( i );
				break;
			}
			++i;
		}
		if (totMin == 0)
			totMin = min;
		for( ++i; i <= nBins; ++i ) {
			double value = histos[j]->GetBinContent( i );
			if( value > 0 && value < min ) {
				min = value;
			}
		}
		if (min > 0 && min < totMin)
			totMin = min;
	}
	return totMin;
}

double findHistosMaximum( const vector<TH1D *> histos, THStack & stack ) {
	double totMax = 0;
	double max;
	for (unsigned j = 0; j < histos.size(); ++j) {
		max = histos[j]->GetMaximum();
		if (max > totMax)
			totMax = max;
	}
	if (stack.GetStack() && stack.GetMaximum() > totMax)
		totMax = stack.GetMaximum();
	return totMax;
}
