#include "samples.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <sstream>
#include <iomanip>
#include <cmath>

using std::ifstream;
using std::cout;
using std::endl;
using std::stringstream;
using std::setw;
using std::setprecision;

Samples::Samples( const char * fileName )
{
	dataSamples.clear();
	signalSamples.clear();
	backgroundSamples.clear();
	workingSignalSample = -1;
	workingDataSample = -1;
	workingLumi = 0;
	readSamples( fileName );
}

Samples::~Samples()
{
}

void Samples::readSamples( const char * fileName )
{
	ifstream input( fileName );
	if( !input.is_open() ) {
		cout << "ERROR: Can't open file with samples (" << fileName << ")!" << endl;
		exit( EXIT_FAILURE );
	}
	while( !input.eof() ) {
		string temp;
		getline( input, temp );
		if( !temp.size() || temp[0] == '\n' || temp[0] == '#' ) {
			continue;
		}

		stringstream in( temp );
		char formatTy;
		char sampTy;
		string name;
		string path;

		if( temp[0] == 'C' ) {
			double xsection;
			unsigned int n_events;
			in >> formatTy >> sampTy >> name >> path >> xsection >> n_events;
			if( sampTy == 'S' )
				signalSamples.push_back( sample( name, path, xsection, n_events, SIGNAL ) );
			else if( sampTy == 'B' )
				backgroundSamples.push_back( sample( name, path, xsection, n_events, BACKGROUND ) );
			else if( sampTy == 'D' )
				dataSamples.push_back( sample( name, path, xsection, n_events, DATA ) );
			else
				cout << "!!! WARNING: WRONG SAMPLE TYPE !!!" << endl;
		} else if( temp[0] == 'L' ) {
			double lumi;
			in >> formatTy >> sampTy >> name >> path >> lumi;
			if( sampTy == 'S' )
				signalSamples.push_back( sample( name, path, lumi, SIGNAL ) );
			else if( sampTy == 'B' )
				backgroundSamples.push_back( sample( name, path, lumi, BACKGROUND ) );
			else if( sampTy == 'D' )
				dataSamples.push_back( sample( name, path, lumi, DATA ) );
			else
				cout << "!!! WARNING: WRONG SAMPLE TYPE !!!" << endl;
		} else
			cout << "!!! WARNING: WRONG FORMAT !!!" << endl;
	}
	if( signalSamples.size() || backgroundSamples.size() || dataSamples.size() )
		cout << "Following samples are read in: " << endl;
	else {
		cout << "ERROR: NO SAMPLES READ!" << endl;
		exit( EXIT_FAILURE );
	}
	if( signalSamples.size() ) {
		cout << "SIGNAL SAMPLES" << endl;
		setWorkingSignalSample( 0 );
	}
	for( unsigned i = 0; i < signalSamples.size(); ++i ) {
		cout << "Sample name: " << setw( 20 ) << signalSamples[i].GetName();
		if( workingSignalSample == (int) i )
			cout << '*';
		cout << ", Xsection: " << setw( 10 ) << signalSamples[i].GetCrossSection()
		     << ", # of events: " << setw( 10 ) << signalSamples[i].GetNevents() << setw( 10 ) << setprecision( 4 ) << signalSamples[i].GetLumi() << endl;
	}
	if( backgroundSamples.size() ) {
		cout << "BACKGROUND SAMPLES" << endl;
	}
	for( unsigned i = 0; i < backgroundSamples.size(); ++i ) {
		cout << "Sample name: " << setw( 20 ) << backgroundSamples[i].GetName() << ", Xsection: " << setw( 10 ) << backgroundSamples[i].GetCrossSection()
		     << ", # of events: " << setw( 10 ) << backgroundSamples[i].GetNevents() << setw( 10 ) << setprecision( 4 ) << backgroundSamples[i].GetLumi()  << endl;
	}
	if( dataSamples.size() ) {
		cout << "DATA SAMPLES" << endl;
		setWorkingDataSample( 0 );
	}
	for( unsigned i = 0; i < dataSamples.size(); ++i ) {
		cout << "Sample name: " << setw( 20 ) << dataSamples[i].GetName();
		if( workingSignalSample == (int) i )
			cout << '*';
		cout << ", Xsection: " << setw( 10 ) << dataSamples[i].GetCrossSection()
		     << ", # of events: " << setw( 10 ) << dataSamples[i].GetNevents() << setw( 10 ) << setprecision( 4 ) << dataSamples[i].GetLumi()  << endl;
	}
}

void Samples::printNumbers( ostream& output, const string& rowName )
{
	output.precision( 4 );
	output << setw( 15 ) << rowName;
	double n_background = 0;
	for( unsigned i = 0; i < backgroundSamples.size(); ++i ) {
		double normEv = backgroundSamples[i].GetNselected() * getBackgroundScaleFactor(i);
		output << setw( 10 ) << backgroundSamples[i].GetNselected() << "," << setw( 9 ) << normEv;
		n_background += normEv;
	}
	output << setw( 10 ) << n_background;
	if( workingDataSample != -1 ) {
		double dataEv = dataSamples[workingDataSample].GetNselected();
		output << setw( 10 ) << dataEv;
	} else
		output << setw( 10 ) << "---";
	if( workingSignalSample != -1 ) {
		double normSigEv = signalSamples[workingSignalSample].GetNselected() * getSignalScaleFactor();
		output << setw( 10 ) << signalSamples[workingSignalSample].GetNselected() << ',' << setw( 9 ) << normSigEv;
		output << setw( 10 ) << significance() << endl;
	} else
		output << setw( 11 ) << "---," << setw( 9 ) << "---" << setw( 10 ) << "---";
	output << endl;
}

void Samples::printNumbersHeading( ostream & output )
{
	output.precision( 4 );
	output << setw( 15 ) << "Cut";
	for( unsigned i = 0; i < backgroundSamples.size(); ++i )
		output << setw( 20 ) << backgroundSamples[i].GetName();
	output << setw( 10 ) << "# BG Ev";
	if( workingDataSample != -1 )
		output << setw( 10 ) << dataSamples[workingDataSample].GetName();
	else
		output << setw( 10 ) << "DATA";
	if( workingSignalSample != -1 )
		output << setw( 20 ) << signalSamples[workingSignalSample].GetName();
	else
		output << setw( 20 ) << "SIGNAL";
	output << setw( 10 ) << "Sign" << endl;
}

void Samples::printScaleFactors( ostream & output )
{
	output.precision( 4 );
	output << setw( 15 ) << "Scale factor";
	for( unsigned i = 0; i < backgroundSamples.size(); ++i )
		output << setw( 20 ) << getBackgroundScaleFactor(i);
	if( workingDataSample != -1 )
		output << setw( 20 ) << workingLumi / dataSamples[workingDataSample].GetLumi();
	else
		output << setw( 20 ) << "---";
	if( workingSignalSample != -1 )
		output << setw( 20 ) << getSignalScaleFactor();
	else
		output << setw( 20 ) << "---";
	output << setw( 10 ) << "---" << endl;
}

void Samples::printRawNumbers( ostream & output )
{
	output.precision( 4 );
	output << setw( 15 ) << "Raw";
	double n_background = 0;
	for( unsigned i = 0; i < backgroundSamples.size(); ++i ) {
		double normEv = backgroundSamples[i].GetNevents() * getBackgroundScaleFactor(i);
		output << setw( 10 ) << backgroundSamples[i].GetNevents() << "," << setw( 9 ) << normEv;
		n_background += normEv;
	}
	output << setw( 10 ) << n_background;
	if( workingDataSample != -1 )
		output << setw( 10 ) << dataSamples[workingDataSample].GetNevents();
	else
		output << setw( 10 ) << "---";
	if( workingSignalSample != -1 ) {
		double n_signal = signalSamples[workingSignalSample].GetNevents() * getSignalScaleFactor();
		output << setw( 10 ) << signalSamples[workingSignalSample].GetNevents() << "," << setw( 9 ) << n_signal;
		double sign = -1;
		if( n_background )
			sign = double( n_signal ) / std::sqrt( n_background + n_signal );
		output << setw( 10 ) << sign;
	} else
		output << setw( 11 ) << "---," << setw( 9 ) << "---" << setw( 10 ) << "---";
	output << endl;
}

void Samples::setWorkingSignalSample( unsigned n )
{
	if( n >= 0 && n < signalSamples.size() )
		workingSignalSample = n;
	else {
		cout << "ERROR: WRONG SIGNAL SAMPLE!" << endl;
		exit( EXIT_FAILURE );
	}
}

void Samples::setWorkingDataSample( unsigned n )
{
	if( n >= 0 && n < dataSamples.size() ) {
		workingDataSample = n;
		workingLumi = dataSamples[workingDataSample].GetLumi();
	} else {
		cout << "ERROR: WRONG DATA SAMPLE!" << endl;
		exit( EXIT_FAILURE );
	}
}

double Samples::significance()
{
	if( workingSignalSample != - 1 ) {
		double n_signal = 0;
		double n_background = 0;
		for( unsigned i = 0; i < backgroundSamples.size(); ++i ) {
			double normEv = backgroundSamples[i].GetNselected() * getBackgroundScaleFactor(i);
			n_background += normEv;
		}
		n_signal = signalSamples[workingSignalSample].GetNselected() * getSignalScaleFactor();
		double sign = -1;
		if( n_background )
			sign = double( n_signal ) / std::sqrt( n_background + n_signal );
		return sign;
	} else
		return -1;
}

void Samples::setWorkingLumi( double l )
{
	if( l > 0 )
		workingLumi = l;
	else {
		cout << "ERROR: WRONG LUMINOSITY!" << endl;
		exit( EXIT_FAILURE );
	}
}

void Samples::selectSampleEvents( sample& samp, const Cut & cut, bool readListFromTree )
{
	TFile * file = new TFile( samp.GetPath().c_str() );
	TTree * tree = ( TTree * ) file->Get( "HZZ4LeptonsAnalysis" );
	event ev( tree );
	TEventList out_list;
	out_list.Reset();
	unsigned int selected = 0;
	if( readListFromTree ) {
		unsigned long nentries = tree->GetEntries();
		for( unsigned long j = 0; j < nentries; j++ ) {
			tree->GetEntry( j );
			if( !cut( ev ) )
				continue;
			selected++;
			out_list.Enter( j );
		}
	} else {
		if( samp.GetEventList().GetN() == 0 )
			cout << "  " << samp.GetName() << ": Empty list!" << endl;
		TEventList in_list( samp.GetEventList() );
		unsigned long nentries = in_list.GetN();
		for( unsigned long j = 0; j < nentries; j++ ) {
			unsigned long en = in_list.GetEntry( j );
			tree->GetEntry( en );
			if( !cut( ev ) )
				continue;
			selected++;
			out_list.Enter( en );
		}
	}
	samp.SetNselected(selected);
	samp.GetEventList() = out_list;
	tree->Delete();
	file->Close();
	delete file;
}

void Samples::selectEvents(const Cut& cut, bool readListFromTree)
{
	for (unsigned i = 0; i < backgroundSamples.size(); ++i)
		selectSampleEvents(backgroundSamples[i], cut, readListFromTree);
	for (unsigned i = 0; i < signalSamples.size(); ++i)
		selectSampleEvents(signalSamples[i], cut, readListFromTree);
	for (unsigned i = 0; i < dataSamples.size(); ++i)
		selectSampleEvents(dataSamples[i], cut, readListFromTree);
}

const Samples::sample & Samples::getBackgroundSample( unsigned n ) {
	if ( n >= backgroundSamples.size() ) {
		cout << "ERROR: Index out of range!" << endl;
		exit( EXIT_FAILURE );
	} else
		return backgroundSamples[n];
}

double Samples::getSignalScaleFactor() {
	return workingLumi / signalSamples[workingSignalSample].GetLumi();
}

double Samples::getBackgroundScaleFactor(unsigned n) {
	if ( n >= backgroundSamples.size() ) {
		cout << "ERROR: Index out of range!" << endl;
		exit( EXIT_FAILURE );
	} else
		return workingLumi / backgroundSamples[n].GetLumi();
}

